!    COLOR FUNCTIONS (routines for Colorhelix & TestRGBtable)
!    Copyright (C) 2011,2012 by Ingo Thies
!
!    Color Functions is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 2 of the License, or
!    (at your option) any later version.
!
!    Color Functions is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with Color Functions. If not, see <http://www.gnu.org/licenses/>.
!
MODULE color_functions_mathconst
  DOUBLE PRECISION :: pi6=0.52359877559829887308d0
  DOUBLE PRECISION :: pi4=0.78539816339744830962d0
  DOUBLE PRECISION :: pi3=1.04719755119659774615d0
  DOUBLE PRECISION :: pi2=1.57079632679489661923d0
  DOUBLE PRECISION :: pi =3.14159265358979323846d0
  DOUBLE PRECISION :: tau=6.28318530717958647693d0
  DOUBLE PRECISION :: r2d=5.729577951308232088d1
  DOUBLE PRECISION :: d2r=1.745329251994329577d-2
END MODULE color_functions_mathconst

!     Matrices are of dimention 3x3 of the following form:
!     /a11 a12 a13\   /array(1,1) array(2,1) array(3,1)\
!     |a21 a22 a23| = |array(1,2) array(2,2) array(3,2)|
!     \a31 a32 a33/   \array(1,3) array(2,3) array(3,3)/
!
!     Note that array(j,i) = aij (left entry revolves fastest)

MODULE color_functions
  INTEGER :: cf_verbose=0
  INTEGER,PARAMETER :: CF_SINGLE=4
  INTEGER,PARAMETER :: CF_DOUBLE=8
! from http://en.wikipedia.org/wiki/LMS_color_space (02.09.2011)
! and http://www.babelcolor.com/download/A%20review%20of%20RGB%20color%20spaces.pdf
  REAL(KIND=CF_DOUBLE) :: krgb2xyz=0.17697d0
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_d50=(/0.96422d0,1.00000d0,0.82521d0/)!sum=2.78943
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_d55=(/0.95682d0,1.00000d0,0.92149d0/)!sum=2.87831
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_d65=(/0.95047d0,1.00000d0,1.08883d0/)!sum=3.03930
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_d75=(/0.94972d0,1.00000d0,1.22638d0/)!sum=3.17610
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_9300k=(/0.97135d0,1.d0,1.43929d0/)!sum=3.41064
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: whitepoint_e=(/1.d0,1.d0,1.d0/)!sum=3.
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: srgb_red=(/0.64d0,0.33d0,0.03d0/)
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: srgb_grn=(/0.30d0,0.60d0,0.10d0/)
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: srgb_blu=(/0.15d0,0.06d0,0.79d0/)
  REAL(KIND=CF_DOUBLE),DIMENSION(3) :: srgb_wht=(/0.95047d0,1.00000d0,1.08883d0/)/3.03930d0
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: srgb2xyz=RESHAPE((/0.4124d0, 0.3576d0, 0.1805d0, &
                                              0.2126d0, 0.7152d0, 0.0722d0,&
                                              0.0193d0, 0.1192d0, 0.9505d0/),(/3,3/))
! from http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: srgb3xyz=RESHAPE((/0.4124564d0, 0.3575761d0, 0.1804375D0,&
                                                        0.2126729d0, 0.7151522d0, 0.0721750D0,&
                                                        0.0193339d0, 0.1191920d0, 0.9503041D0/),(/3,3/))
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: widergb2xyz=RESHAPE((/0.7161046d0, 0.1009296d0, 0.1471858d0,&
                                                           0.2581874d0, 0.7249378d0, 0.0168748d0,&
                                                           0.0000000d0, 0.0517813d0, 0.7734287d0/),(/3,3/))
! from http://en.wikipedia.org/wiki/LMS_color_space (02.09.2011)
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: srgbwp=RESHAPE((/3.2406d0,-1.5372d0,-0.4986d0,&
                                                   -0.9689d0, 1.8758d0, 0.0415d0,&
                                                    0.0557d0,-0.2040d0, 1.0570d0/),(/3,3/))
! from http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: srgb3=RESHAPE((/ 3.2404542d0,-1.5371385d0,-0.4985314d0,&
                                                     -0.9692660d0, 1.8760108d0, 0.0415560d0,&
                                                      0.0556434d0,-0.2040259d0, 1.0572252d0/),(/3,3/))
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: widergb=RESHAPE((/ 1.4628067d0,-0.1840623d0,-0.2743606d0,&
                                                       -0.5217933d0, 1.4472381d0, 0.0677227d0,&
                                                        0.0349342d0,-0.0968930d0, 1.2884099d0/),(/3,3/))
! from http://www.w3.org/Graphics/Color/sRGB
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: w3srgb=RESHAPE((/3.2410d0,-1.5374d0,-0.4986d0,&
                                                   -0.9692d0, 1.8760d0, 0.0416d0,&
                                                    0.0556d0,-0.2040d0, 1.0570d0/),(/3,3/))
! from own evaluation based on D65 white point XYZ = 0.95047, 1.00000, 1.08883
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: xrgb,xrgb2xyz!dynamic evaluation
!---- Chromatic adation cone response domains (www.brucelindbloom.com)
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: ma_bradford=RESHAPE((/0.8951000d0, 0.2664000d0,-0.1614000d0,&
                                                          -0.7502000d0, 1.7135000d0, 0.0367000d0,&
                                                           0.0389000d0,-0.0685000d0, 1.0296000d0/),(/3,3/))
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: ma_vonkries=RESHAPE((/0.4002400d0, 0.7076000d0,-0.0808100d0,&
                                                          -0.2263000d0, 1.1653200d0, 0.0457000d0,&
                                                           0.0000000d0, 0.0000000d0, 0.9182200d0/),(/3,3/))
  REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: mainv_bradford,mainv_vonkries

  CONTAINS
    
    SUBROUTINE mapto01(rgb,rgb01,bordermode,crange)
! Trim linear RGB to within [0,1]
! floormode   = left digit
! ceilmode    = right digit (i. bordermode=10 means floormode=1, ceilmode=0)
!            <0 just truncate outside 0 and 1
! floor:      0 scale to c0 = 0
!             1 scale to c0 = minval(rgb)
!             2 scale to c0 = min(minval(rgb),crange(0)) (previous default for crange(0)=0)
!             3 scale to c0 = crange(0) (use =0 for color chart)
! ceiling:    0 scale to c1 = 1
!             1 scale to c1 = maxval(rgb) (the previous default)
!             2 scale to c1 = max(maxval(rgb),crange(1)) 
!             3 scale to c1 = crange(1)
! bordermode -1 don't scale anything, just clip
!           <-1 don't even clip (not recommended for general use!)
! Example: For vivid colorcharts (at least one channel maximised) use bordermode=31
      IMPLICIT NONE
      INTEGER bordermode,floormode,ceilmode,j
      REAL(KIND=CF_DOUBLE) :: rgb(1:3),rgb01(1:3),crange(0:1),c0,c1,scalrgb
      REAL(KIND=CF_DOUBLE) :: grey,gdev,yerg,drgbmin
      REAL(KIND=CF_DOUBLE) :: srgb(1:3)
      IF (bordermode<0) THEN
         IF (bordermode==-1) THEN
            DO j=1,3
               rgb01(j)=max(min(rgb(j),1.d0),0.d0)!just clip
            ENDDO
         ELSE
            rgb01=rgb
         ENDIF
         RETURN
      ENDIF
      floormode=bordermode/10
      ceilmode=bordermode-10*floormode
      IF (floormode==1) THEN
         c0=minval(rgb)
      ELSE IF (floormode==2) THEN
         c0=min(minval(rgb),crange(0))
      ELSE IF (floormode==3) THEN
         c0=crange(0)
      ELSE
         c0=0.d0
      ENDIF
      IF (ceilmode==1) THEN
!  Full value of at least one channel
         c1=maxval(rgb)
      ELSE IF (ceilmode==2) THEN
!  Allow lower values
         c1=max(maxval(rgb),crange(1))
      ELSE IF (ceilmode==3) THEN
         c1=crange(1)
      ELSE IF (ceilmode==4) THEN
!**** Special mode to maximize grey distance *****
         grey=crange(0)!grey axis
         gdev=max(min(crange(1),1.d0),0.d0)!scaling for saturation
         yerg=1.d0-grey
         DO j=1,3
            IF (rgb(j)>=grey) THEN!distance to ceiling
               srgb(j)=yerg/max(rgb(j)-grey,1.d-12)
            ELSE!distance to floor
               srgb(j)=grey/max(grey-rgb(j),1.d-12)
            ENDIF
         ENDDO
         scalrgb=minval(srgb)*gdev
         rgb01(:)=grey+(scalrgb*(rgb(:)-grey))
!---- Exit here
         RETURN
!*************************************************
      ELSE
         c1=1.d0
      ENDIF
      scalrgb=c1-c0
      DO j=1,3
         rgb01(j)=max(min((rgb(j)-c0)/(c1-c0),1.d0),0.d0)
      ENDDO
    END SUBROUTINE mapto01

    SUBROUTINE desaturate_rgb(rgbin,rgbout,fwhite)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: rgbin(1:3),rgbout(1:3),rgbadd(1:3),fwhite,fcol=0.d0,fcomp
      LOGICAL :: addblack
      addblack=fwhite<0.d0
      IF (addblack) THEN
         fwhite=-fwhite
         rgbadd=(/0.d0,0.d0,0.d0/)
      ELSE
         rgbadd=(/1.d0,1.d0,1.d0/)
      ENDIF
      fcol=1.d0-fwhite
      rgbout(:)=fcol*rgbin(:)+fwhite*rgbadd(:)
      RETURN
    END SUBROUTINE desaturate_rgb

!===============================================================================

  SUBROUTINE wrap_rgb_gamma(Jin,Jout,gamma,direction)
    IMPLICIT NONE
    INTEGER :: direction
    REAL(KIND=CF_DOUBLE) :: gamma
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Jin,Jout
    IF (gamma<1.d-2) THEN!Assume sRGB if gamma is tiny
       CALL srgb_gamma(Jin,Jout,direction)
    ELSE
       CALL corr_gamma(Jin,Jout,gamma,direction)
    ENDIF
    RETURN
  END SUBROUTINE wrap_rgb_gamma

    SUBROUTINE corr_gamma(Jin,Jout,gamma,direction)
      IMPLICIT NONE
      INTEGER :: direction,i
      REAL(KIND=CF_DOUBLE) :: Jin(3),Jout(3),gamma,clin,c
!      write(*,'(a,f6.3,2x,i2)') 'gamma,direction=',gamma,direction
!      write(*,'(a,6(e11.4))')    'Jin            =',Jin
      IF (direction>=1) THEN
! RGB_lin to RGB_gamma
         DO i=1,3
            clin=Jin(i)
            IF (clin>0.d0) THEN
               c=clin**(1.d0/gamma)
!               write(*,'(a,6(e11.4))') ' >0: clin,c,gamma =',clin,c,gamma
            ELSE
!               write(*,'(a,6(e11.4))') '<=0: clin,c,gamma =',clin,c,gamma
               c=0.d0!handle zero separately
            ENDIF
            Jout(i)=c
         ENDDO
      ELSE IF (direction<=-1) THEN
! RGB_gamma to RGB_lin
         DO i=1,3
            c=Jin(i)
            IF (c>0.d0) THEN
               clin=c**gamma
!               write(*,'(a,6(e11.4))') '-1: clin,c,gamma =',clin,c,gamma
            ELSE
               clin=0.d0!handle zero separately
            ENDIF
            Jout(i)=clin
         ENDDO
      ELSE
! No gamma correction
         Jout=Jin
      ENDIF
!      write(*,'(a,6(e11.4)/)') 'Jout        =',Jout
      RETURN
    END SUBROUTINE corr_gamma

    SUBROUTINE srgb_gamma(Jin,Jout,direction)
      IMPLICIT NONE
      INTEGER :: direction,i
      REAL(KIND=CF_DOUBLE) :: Jin(3),Jout(3)
      REAL(KIND=CF_DOUBLE) :: Jrgblin(3),Jsrgb(3),c,clin
      REAL(KIND=CF_DOUBLE) :: a=5.5d-2
      IF (direction>=1) THEN
! RGB_lin to RGB_gamma
         DO i=1,3
            clin=Jin(i)
            IF (clin<=0.0031308d0) THEN
               c=12.92d0*clin
            ELSE
               c=(1.d0+a)*clin**(1.d0/2.4d0)-a
            ENDIF
            Jout(i)=c
         ENDDO
      ELSE IF (direction<=-1) THEN
! RGB_gamma to RGB_lin
         DO i=1,3
            c=Jin(i)
            IF (c<=0.04045d0) THEN
               clin=c/12.92d0
            ELSE
               clin=((c+a)/(1.d0+a))**2.4d0
            ENDIF
            Jout(i)=clin
         ENDDO
      ELSE
! No gamma correction
         Jout=Jin
      ENDIF
      RETURN
    END SUBROUTINE srgb_gamma

    SUBROUTINE rescale_rgb(rgbi,rgbo,yold,ynew,gamma,mode)
      IMPLICIT NONE
      INTEGER :: mode
      DOUBLE PRECISION :: rgbi(3),rgbo(3),yold,ynew,gamma,qy,qintens
      DOUBLE PRECISION :: rgblin(3),rgbdum1(3),rgbdum2(3)
      IF (mode==0) THEN
         rgbo=rgbi
         RETURN
      ELSE IF (mode<0) THEN!treat yold and ynew as RGB levels
         IF (ynew<=yold) THEN
            qy=ynew/yold
         ELSE
            qy=yold/ynew
         ENDIF
         rgbdum1=(/0.d0,qy,0.d0/)
         CALL wrap_rgb_gamma(rgbdum1,rgbdum2,gamma,-1)
         IF (ynew<=yold) THEN
            qintens=rgbdum2(2)
         ELSE
            qintens=1.d0/rgbdum2(2)
         ENDIF
      ELSE!assume yold and ynew as Y levels (i.e. linear)
         qintens=ynew/yold
      ENDIF
!         CALL srgb_gamma(rgbi,rgblin,-1)
      CALL wrap_rgb_gamma(rgbi,rgblin,gamma,-1)
      rgblin(:)=rgblin(:)*qintens
!         CALL srgb_gamma(rgblin,rgbo,1)
      CALL wrap_rgb_gamma(rgblin,rgbo,gamma,1)
      RETURN
    END SUBROUTINE rescale_rgb

!===============================================================================

  SUBROUTINE xyz2uv35(Jxyz,uv35)
! Judd (1935) uniform chromaticity space
    IMPLICIT NONE
    INTEGER :: i
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Jxyz,xyz
    REAL(KIND=CF_DOUBLE),DIMENSION(2) :: uv35
    REAL(KIND=CF_DOUBLE) :: x,y,u,v,div,sumxyz
    REAL(KIND=CF_DOUBLE) :: a=0.4661d0,b=0.1593d0,c=-0.15735d0,d=0.2424d0,e=0.6581d0
    sumxyz=sum(Jxyz)
    DO i=1,3
       xyz(i)=Jxyz(i)/sumxyz
    ENDDO
    x=xyz(1); y=xyz(2)
    div=y+c*x+d
    u=(a*x+b*y)/div
    v=e*y/div
    uv35=(/u,v/)
    RETURN
  END SUBROUTINE xyz2uv35

  SUBROUTINE uv2xyz35(uv35,xyz)
! Own backward-transformation
    IMPLICIT NONE
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: xyz
    REAL(KIND=CF_DOUBLE),DIMENSION(2) :: uv35
    REAL(KIND=CF_DOUBLE) :: x,y,u,v,umb,vme,cuma,cv,du,dv
    REAL(KIND=CF_DOUBLE) :: a=0.4661d0,b=0.1593d0,c=-0.15735d0,d=0.2424d0,e=0.6581d0
    u=uv35(1); v=uv35(2)
    umb=u-b
    vme=v-e
    cuma=c*u-a
    cv=c*v
    du=d*u
    dv=d*v
    x=(umb*dv-vme*du)/(cuma*vme-cv*umb)
    y=(cuma*dv-cv*du)/(cv*umb-cuma*vme)
    xyz=(/x,y,1.d0-x-y/)
    RETURN
  END SUBROUTINE uv2xyz35

    SUBROUTINE xyz2uv(Jxyz,uv)
! CIE1931 XYZ to CIE 1960 uv
      IMPLICIT NONE
      INTEGER i
      REAL(KIND=CF_DOUBLE) :: Jxyz(3),uv(2),xyz(3),sumxyz,divxy,x,y
      sumxyz=sum(Jxyz)
      DO i=1,3
         xyz(i)=Jxyz(i)/sumxyz
      ENDDO
      x=xyz(1); y=xyz(2)
      divxy=(-2.d0*x+12.d0*y+3.d0)
      uv(1)=4.d0*x/divxy
      uv(2)=6.d0*y/divxy
      RETURN
    END SUBROUTINE xyz2uv

    SUBROUTINE uv2xyz(uv,Jxyz)
! CIE1960 uv to CIE1931 xyz (constant lightness)
      IMPLICIT NONE
      INTEGER i
      REAL(KIND=CF_DOUBLE) :: Jxyz(3),uv(2),xyz(3),sumxyz,divuv,u,v
!      sumxyz=sum(Jxyz)
!      DO i=1,3
!         xyz(i)=Jxyz(i)/sumxyz
!      ENDDO
      u=uv(1); v=uv(2)
      divuv=(2.d0*u-8.d0*v+4.d0)
      xyz(1)=3.d0*u/divuv
      xyz(2)=2.d0*v/divuv
      xyz(3)=1.d0-xyz(1)-xyz(2)
      Jxyz(:)=xyz(:)
      RETURN
    END SUBROUTINE uv2xyz

    SUBROUTINE xyz2uv76(Jxyz,uv76)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: Jxyz(3),uv76(2),uv(2)
      CALL xyz2uv(Jxyz,uv)
      uv76(1)=uv(1)
      uv76(2)=uv(2)*1.5d0
      RETURN
    END SUBROUTINE xyz2uv76

    SUBROUTINE uv2xyz76(uv76,Jxyz)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: Jxyz(3),uv76(2),uv(2)
      uv(1)=uv76(1)
      uv(2)=uv76(2)/1.5d0
      CALL uv2xyz(uv,Jxyz)
      RETURN
    END SUBROUTINE uv2xyz76

    SUBROUTINE uvy2uvw64(uv,y100,uvn,uvw100)
! xyz100 means xyz~100 = 100*Jxyz
! For illuminants use Y=xyz(2)=100.
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: uv(2),y100,uvn(2),uvw100(3)
      REAL(KIND=CF_DOUBLE) :: ustar,vstar,wstar,third=1.d0/3.d0
      wstar=25.d0*y100**third-17.d0
      ustar=13.d0*wstar*(uv(1)-uvn(1))
      vstar=13.d0*wstar*(uv(2)-uvn(2))
      uvw100=(/ustar,vstar,wstar/)
      RETURN
    END SUBROUTINE uvy2uvw64

    SUBROUTINE xyz2uvw64(xyz100,xyzwhite,uvw100)
! xyz100 means xyz~100 = 100*Jxyz
! For illuminants use Y=xyz(2)=100.
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: xyz100(3),xyzwhite(3),uvw100(3)
      REAL(KIND=CF_DOUBLE) :: ustar,vstar,wstar,third=1.d0/3.d0
      REAL(KIND=CF_DOUBLE) :: y100,uv(2),uvn(2),sumxyz,divuv
      y100=xyz100(2)
      CALL xyz2uv(xyz100,uv)
      CALL xyz2uv(xyzwhite,uvn)
      CALL uvy2uvw64(uv,y100,uvn,uvw100)
!      wstar=25.d0*y100**third-17.d0
!      ustar=13.d0*wstar*(uv(1)-uvn(1))
!      vstar=13.d0*wstar*(uv(2)-uvn(2))
!      uvw100=(/ustar,vstar,wstar/)
      RETURN
    END SUBROUTINE xyz2uvw64

    SUBROUTINE xyz2lab(XYZ,XYZwhite,Lab)
! XYZ assumed to be absolute (i.e. Y = luminance)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: Lab(3),XYZ(3),XYZwhite(3),L16
      REAL(KIND=CF_DOUBLE) :: L,a,b,xr,yr,zr,fx,fy,fz
      xr=XYZ(1)/XYZwhite(1)
      yr=XYZ(2)/XYZwhite(2)
      zr=XYZ(3)/XYZwhite(3)
      fx=flab_fwd(xr)
      fy=flab_fwd(yr)
      fz=flab_fwd(zr)
      L=116.d0*fy-16.d0
      a=500.d0*(fx-fy)
      b=200.d0*(fy-fz)
      Lab=(/L,a,b/)
      RETURN
    END SUBROUTINE xyz2lab

    SUBROUTINE lab2xyz(Lab,XYZ,XYZwhite)
! Input:
!     Lab(1:3)    - Vector that contains the L,a,b data
! Output:
!     XYZ(1:3)    - Vector that contains the X,Y,Z data
! Parameter:
!     XYZwhite    - Vector that contains the whitepoint X,Y,Z
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: Lab(3),XYZ(3),XYZwhite(3),L16
      REAL(KIND=CF_DOUBLE) :: L,a,b
      L=Lab(1);a=Lab(2);b=Lab(3)
      L16=(L+16.d0)/116.d0
      XYZ(1)=XYZwhite(1)*flab_bwd(L16+a/500.d0)
      XYZ(2)=XYZwhite(2)*flab_bwd(L16)
      XYZ(3)=XYZwhite(3)*flab_bwd(L16-b/200.d0)!sign corrected here
!      XYZ(3)=XYZwhite(3)*flab_bwd(L16+b/200.d0)!old wrong formula
!      write(*,*) 'x,y,z =',xyz(1),xyz(2),xyz(3)
      RETURN
    END SUBROUTINE lab2xyz

    FUNCTION flab_fwd(t)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: flab_fwd,t
      REAL(KIND=CF_DOUBLE) :: c1=6.d0/29.d0,c2=4.d0/29.d0,third=1.d0/3.d0
      IF (t>c1**3) THEN
         flab_fwd=t**third
      ELSE
         flab_fwd=t/(3.d0*c1**2)+c2
      ENDIF
    END FUNCTION flab_fwd

    FUNCTION flab_bwd(t)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: flab_bwd,t
      REAL(KIND=CF_DOUBLE) :: c1=6.d0/29.d0,c2=4.d0/29.d0
      IF (t>c1) THEN
         flab_bwd=t**3
      ELSE
         flab_bwd=3.d0*c1**2*(t-c2)
      ENDIF
    END FUNCTION flab_bwd


!===============================================================================

    SUBROUTINE whitepoint_d(cct,xd,yd,mdata)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: cct,xd,yd,mdata(0:2)
      REAL(KIND=CF_DOUBLE) :: a0=0.244063D0,a1=0.09911D0,a2=2.9678D0,a3=-4.6070D0
      REAL(KIND=CF_DOUBLE) :: b0=0.237040D0,b1=0.24748D0,b2=1.9018D0,b3=-2.0064D0
      REAL(KIND=CF_DOUBLE) :: c0=-0.275D0,c1=2.870D0,c2=-3.000D0
      REAL(KIND=CF_DOUBLE) :: s
!==== Compute chromaticity
      s=1.d3/cct
      IF (cct.le.7.d3) THEN
         xd=a0+(a1+(a2+a3*s)*s)*s
      ELSE
         xd=b0+(b1+(b2+b3*s)*s)*s
      ENDIF
      yd=c0+(c1+c2*xd)*xd
!==== Coefficients
      mdata(0)=0.0241D0+0.2562D0*xd-0.7341D0*yd
      mdata(1)=(-1.3515D0-1.7703D0*xd+5.9114D0*yd)/mdata(0)
      mdata(2)=(0.03000D0-31.4424D0*xd+30.0717D0*yd)/mdata(0)
      RETURN
    END SUBROUTINE whitepoint_d

    SUBROUTINE brightness_correction(Jxyz,f,kf)
!==== Correction factor for brightness for constand luminance
!     as a function of chromaticity
!     (Kaiser 1986, retrieved from Ohta & Robertsen,
!     Colorimetry - Fundamentals and Applications,
!     2005, equation 1.40)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE) :: Jxyz(3),x,y,f,kf,scalxyz
      scalxyz=sum(Jxyz)
      x=Jxyz(1)/scalxyz; y=Jxyz(2)/scalxyz
      f=0.256d0-0.184d0*y-2.527d0*x*y+4.656d0*x**3*y+4.657d0*x*y**4
      kf=1.d1**f
      RETURN
    END SUBROUTINE brightness_correction

!===============================================================================

    SUBROUTINE operate_matrix3x3(mat,vec0,vec1)
!     vec1 = mat vec0
!     mat is a 3x3 matrix of the following form:
!     /a11 a12 a13\   /mat(1,1) mat(2,1) mat(3,1)\
!     |a21 a22 a23| = |mat(1,2) mat(2,2) mat(3,2)|
!     \a31 a32 a33/   \mat(1,3) mat(2,3) mat(3,3)/
!
!     Note that mat(j,i) = aij (left entry revolves fastest)
      IMPLICIT NONE
      INTEGER i,j
      REAL(KIND=CF_DOUBLE) mat(3,3),vec0(3),vec1(3)
      vec1(1)=0.d0
      vec1(2)=0.d0
      vec1(3)=0.d0
      DO i=1,3
         DO j=1,3
            vec1(i)=vec1(i)+mat(j,i)*vec0(j)!inverse order of i,j for optimum memory acces
         ENDDO
      ENDDO
      RETURN
    END SUBROUTINE operate_matrix3x3

    SUBROUTINE invert_matrix3x3(mat,tam,det)
      IMPLICIT NONE
      INTEGER i,j,ii
      REAL(KIND=CF_DOUBLE) mat(3,3),tam(3,3),det
      det=0.d0
      det=  mat(1,1)*mat(2,2)*mat(3,3) - mat(3,1)*mat(2,2)*mat(1,3)  &
           + mat(2,1)*mat(3,2)*mat(1,3) - mat(2,1)*mat(1,2)*mat(3,3) &
           + mat(3,1)*mat(1,2)*mat(2,3) - mat(1,1)*mat(3,2)*mat(2,3)
!     (a  b  c)               (ei-fh  ch-bi  bf-ce)
!     (d  e  f) inv = 1/det * (fg-di  ai-cg  cd-af)
!     (g  h  i)               (dh-eg  bg-ah  ae-bd)
!
!     (a11  b12  c13)               (e22*i33-f23*h32  c13*h32-b12*i33  b12*f23-c13*e22)
!     (d21  e22  f23) inv = 1/det * (f23*g31-d21*i33  a11*i33-c13*g31  c13*d21-a11*f23)
!     (g31  h32  i33)               (d21*h32-e22*g31  b12*g31-a11*h32  a11*e22-b12*d21)
!
!     (a11  a12  a13)               (a22*a33-a23*h32  a13*a32-a12*i33  a12*a23-a13*a22)
!     (a21  a22  a23) inv = 1/det * (a23*a31-a21*i33  a11*a33-a13*g31  a13*a21-a11*a23)
!     (a31  a32  a33)               (a21*a32-a22*g31  a12*a31-a11*h32  a11*a22-a12*a21)
      tam(1,1)=mat(2,2)*mat(3,3)-mat(3,2)*mat(2,3)
      tam(2,1)=mat(3,1)*mat(2,3)-mat(2,1)*mat(3,3)
      tam(3,1)=mat(2,1)*mat(3,2)-mat(3,1)*mat(2,2)
      
      tam(1,2)=mat(3,2)*mat(1,3)-mat(1,2)*mat(3,3)
      tam(2,2)=mat(1,1)*mat(3,3)-mat(3,1)*mat(1,3)
      tam(3,2)=mat(3,1)*mat(1,2)-mat(1,1)*mat(3,2)
      
      tam(1,3)=mat(1,2)*mat(2,3)-mat(2,2)*mat(1,3)
      tam(2,3)=mat(2,1)*mat(1,3)-mat(1,1)*mat(2,3)
      tam(3,3)=mat(1,1)*mat(2,2)-mat(2,1)*mat(1,2)
      DO i=1,3
         DO j=1,3
            tam(j,i)=tam(j,i)/det
         ENDDO
      ENDDO
      RETURN
    END SUBROUTINE invert_matrix3x3

    SUBROUTINE product_matrix3x3(mat1,mat2,mat12)
      IMPLICIT NONE
      INTEGER i,j,k
      REAL(KIND=CF_DOUBLE) mat1(3,3),mat2(3,3),mat12(3,3),sum
      DO i=1,3
         DO j=1,3
            sum=0.d0
            DO k=1,3
               sum=sum+mat1(k,i)*mat2(j,k)
            ENDDO
            mat12(j,i)=sum
         ENDDO
      ENDDO
      RETURN
    END SUBROUTINE product_matrix3x3
          
    SUBROUTINE transpose_matrix3x3(a,atp)
      IMPLICIT NONE
      INTEGER i,j
      REAL(KIND=CF_DOUBLE) a(3,3),atp(3,3)
      DO i=1,3
         DO j=1,3
            atp(j,i)=a(i,j)
         ENDDO
      ENDDO
      RETURN
    END SUBROUTINE transpose_matrix3x3
   
!===============================================================================

    SUBROUTINE getrgb(xyzr,xyzg,xyzb,xyzw,axyz2rgb,argb2xyz)
! XYZw should be scaled to Y=1 for standard matrix
      IMPLICIT NONE
      INTEGER j
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: xyzr,xyzg,xyzb,xyzw
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: testxyzr,testxyzg,testxyzb,testxyzw
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: testr,testg,testb,testw
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: oner,oneg,oneb,onew
      REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: axyz2rgb,argb2xyz
      REAL(KIND=CF_DOUBLE) :: k,c1,c2,c3,adet
      REAL(KIND=CF_DOUBLE) :: a11,a12,a13, a21,a22,a23, a31,a32,a33
!    Set R=0
      k=xyzg(1)/xyzb(1)
!     WRITE(*,*) 'kr1 =',k
      c1=0.d0
      c2=xyzg(2)-k*xyzb(2)
      c3=xyzg(3)-k*xyzb(3)
      a12=-c3/c2
      k=xyzg(2)/xyzb(2)
!     WRITE(*,*) 'kr2 =',k
      c1=xyzg(1)-k*xyzb(1)
      c2=0.d0
      c3=xyzg(3)-k*xyzb(3)
      a11=-c3/c1
!    Set G=0
      k=xyzr(1)/xyzb(1)
!     WRITE(*,*) 'kg1 =',k
      c1=0.d0
      c2=xyzr(2)-k*xyzb(2)
      c3=xyzr(3)-k*xyzb(3)
      a22=-c3/c2
      k=xyzr(2)/xyzb(2)
!     WRITE(*,*) 'kg2 =',k
      c1=xyzr(1)-k*xyzb(1)
      c2=0.d0
      c3=xyzr(3)-k*xyzb(3)
      a21=-c3/c1
!    Set B=0
      k=xyzr(1)/xyzg(1)
!     WRITE(*,*) 'kb1 =',k
      c1=0.d0
      c2=xyzr(2)-k*xyzg(2)
      c3=xyzr(3)-k*xyzg(3)
      a32=-c3/c2
      k=xyzr(2)/xyzg(2)
!     WRITE(*,*) 'kb2 =',k
      c1=xyzr(1)-k*xyzg(1)
      c2=0.d0
      c3=xyzr(3)-k*xyzg(3)
      a31=-c3/c1
!
!
      a13=1.d0/(a11*xyzw(1)+a12*xyzw(2)+xyzw(3))
      a23=1.d0/(a21*xyzw(1)+a22*xyzw(2)+xyzw(3))
      a33=1.d0/(a31*xyzw(1)+a32*xyzw(2)+xyzw(3))
      a11=a11*a13; a12=a12*a13
      a21=a21*a23; a22=a22*a23
      a31=a31*a33; a32=a32*a33
      axyz2rgb=RESHAPE((/a11,a12,a13,&
           a21,a22,a23,&
           a31,a32,a33/),(/3,3/))
      CALL invert_matrix3x3(axyz2rgb,argb2xyz,adet)
      IF (cf_verbose<1) RETURN!skip test output
      WRITE(*,*)
      WRITE(*,'(A)') '************************************************************'
      WRITE(*,'(A)') 'GETRGB: Show Matrices'
      WRITE(*,'(A)') '        XYZ to RGB'
      WRITE(*,'(A,3(1X,F16.12))') 'a11,a12,a13 =',(axyz2rgb(1,j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'a21,a22,a23 =',(axyz2rgb(2,j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'a31,a32,a33 =',(axyz2rgb(3,j),j=1,3)
      WRITE(*,*)
      WRITE(*,'(A)') '        RGB to XYZ'
      WRITE(*,'(A,3(1X,F16.12))') 'b11,b12,b13 =',(argb2xyz(1,j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'b21,b22,b23 =',(argb2xyz(2,j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'b31,b32,b33 =',(argb2xyz(3,j),j=1,3)
      WRITE(*,*)
      WRITE(*,'(A)') 'GETRGB: Test Matrices'
      WRITE(*,'(A)') '        XYZ to RGB'
      testxyzr=xyzr;testxyzg=xyzg;testxyzb=xyzb;testxyzw=xyzw
      CALL operate_matrix3x3(axyz2rgb,testxyzr,testr)
      CALL operate_matrix3x3(axyz2rgb,testxyzg,testg)
      CALL operate_matrix3x3(axyz2rgb,testxyzb,testb)
      CALL operate_matrix3x3(axyz2rgb,testxyzw,testw)
      WRITE(*,'(A,3(1X,F16.12))') 'RGB(R) =',(testr(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'RGB(G) =',(testg(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'RGB(B) =',(testb(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'RGB(W) =',(testw(j),j=1,3)
      CALL operate_matrix3x3(argb2xyz,testr,testxyzr)
      CALL operate_matrix3x3(argb2xyz,testg,testxyzg)
      CALL operate_matrix3x3(argb2xyz,testb,testxyzb)
      CALL operate_matrix3x3(argb2xyz,testw,testxyzw)
      WRITE(*,*)
      WRITE(*,'(A)') '        RGB to XYZ'
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(R) =',(testxyzr(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(G) =',(testxyzg(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(B) =',(testxyzb(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(W) =',(testxyzw(j),j=1,3)
      oner=(/1.d0,0.d0,0.d0/)
      oneg=(/0.d0,1.d0,0.d0/)
      oneb=(/0.d0,0.d0,1.d0/)
      onew=(/1.d0,1.d0,1.d0/)
      CALL operate_matrix3x3(argb2xyz,oner,testxyzr)
      CALL operate_matrix3x3(argb2xyz,oneg,testxyzg)
      CALL operate_matrix3x3(argb2xyz,oneb,testxyzb)
      CALL operate_matrix3x3(argb2xyz,onew,testxyzw)
      WRITE(*,*)
      WRITE(*,'(A)') '        RGB uni to XYZ'
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(R) =',(testxyzr(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(G) =',(testxyzg(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(B) =',(testxyzb(j),j=1,3)
      WRITE(*,'(A,3(1X,F16.12))') 'xyz(W) =',(testxyzw(j),j=1,3)
      WRITE(*,'(A)') '************************************************************'
      WRITE(*,*)
      RETURN
    END SUBROUTINE getrgb

    SUBROUTINE XYZ2srgb(Jxyz,ymode,rgblin,srgb,a2rgb,bordermode,crange)
!     ylum = 0: no luminance scaling
!            1: assume Ymax = 1
!            2: assume Ymax = 100
      IMPLICIT NONE
      INTEGER :: ymode,bordermode
      REAL(KIND=CF_DOUBLE),DIMENSION(3,3) :: a2rgb
      REAL(KIND=CF_DOUBLE),DIMENSION(2) :: crange
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Jxyz,xyz,XYZwhite,rgb0,rgblin,srgb0,srgb
      REAL(KIND=CF_DOUBLE) :: scalxyz,ylum!,x,y,z,xw,yw
      scalxyz=sum(Jxyz)
      IF (abs(ymode)==1) THEN
         ylum=Jxyz(2)
         xyz(:)=Jxyz(:)
!         write(*,*) 'ymode=1: ylum =',ylum
      ELSE IF (ymode==2) THEN
         ylum=Jxyz(2)/1.d2
         xyz(:)=Jxyz(:)/1.d2
!         write(*,*) 'ymode=2: ylum =',ylum
      ELSE
         ylum=1.d0
         xyz(:)=Jxyz(:)/scalxyz
!         write(*,*) 'ymode=0'
      ENDIF
!! Whitepoint
!      IF (xw*yw<=0.d0) THEN
!         XYZwhite(:)=whitepoint_d65(:)
!      ELSE
!         XYZwhite(1)=xw
!         XYZwhite(2)=yw
!         XYZwhite(3)=1.d0-xw-yw
!      ENDIF
      CALL operate_matrix3x3(a2rgb,xyz,rgb0)
      IF (ymode>=0) THEN
!         write(*,*) 'ymode =',ymode
         CALL mapto01(rgb0,rgblin,bordermode,crange)
      ELSE
         rgblin=rgb0
      ENDIF
!      IF (ymode>0) THEN
!         rgb0(:)=rgb0(:)*ylum
!      ENDIF
      CALL srgb_gamma(rgblin,srgb,1)
!      CALL mapto01(rgb0,rgblin,bordermode,crange)
!      CALL mapto01(srgb0,srgb,bordermode,crange)
      RETURN
    END SUBROUTINE XYZ2srgb

  END MODULE color_functions

!===============================================================================
!==== Extended functions (DIN99d+o) ============================================
!===============================================================================

MODULE color_functions_extended
  USE color_functions_mathconst
  USE color_functions
  REAL(KIND=CF_DOUBLE) :: kE_DIN99=1.d0,kCH_DIN99=1.d0

  CONTAINS

    SUBROUTINE lab2lch(Lab,LCh,direction,degmode)
      IMPLICIT NONE
      INTEGER direction,degmode
      REAL(KIND=CF_DOUBLE) :: Lab(3),LCh(3),L,a,b,C,h,xdum
      IF (direction>=1) THEN!L*a*b* to L*C*h*
         L=Lab(1);a=Lab(2);b=Lab(3)
         C=sqrt(a**2+b**2)
         IF (C.gt.0.d0) THEN
            xdum=max(min(a/(C),1.d0),-1.d0)
         ELSE
            xdum=0.d0
         ENDIF
         h=acos(xdum)*sign(1.d0,b)
         IF (h<0.d0) THEN
            h=h+tau
         ELSE IF (h>tau) THEN
            h=h-tau
         ENDIF
         IF (degmode>=1) THEN
            h=h*r2d
         ENDIF
         LCh=(/L,C,h/)
      ELSE! L*C*h* to L*a*b*
         L=LCh(1);C=LCh(2);h=LCh(3)
         IF (degmode>=1) THEN
            h=h/r2d
         ENDIF
         a=C*cos(h)
         b=C*sin(h)
         Lab=(/L,a,b/)
      ENDIF
      RETURN
    END SUBROUTINE lab2lch

!-- DIN99d formulae, retrieved from thesis by Yang Zue (2008) and Shizhe Shen (2009)

    SUBROUTINE XYZ2DIN99d(Ixyz,Ixyzwhite,Lab99d,LCh99d,degmode)
      IMPLICIT NONE
      INTEGER :: degmode
      REAL(KIND=CF_DOUBLE) :: scal_L99d=325.2194339489675351d0
      REAL(KIND=CF_DOUBLE) :: Ix,Iy,Iz,Ls,as,bs,L99d,C99d,h99d,a99d,b99d
      REAL(KIND=CF_DOUBLE) :: e,f,G,hef
      REAL(KIND=CF_DOUBLE) :: d50=0.87266462599716479d0
      REAL(KIND=CF_DOUBLE) :: c50=0.64278760968653933d0
      REAL(KIND=CF_DOUBLE) :: s50=0.76604444311897804d0
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Ixyz,Ixyzwhite,Kxyz,Kxyzwhite,Labs,LCh, &
           Lab99d,LCh99d
!   X modification has to be applied to Whitepoint coordinates, too.
!   Otherwise, colorfulness C99d won't be zero.
      Ix=Ixyzwhite(1);Iy=Ixyzwhite(2);Iz=Ixyzwhite(3)
      Kxyzwhite=(/1.12d0*Ix-0.12d0*Iz,Iy,Iz/)
      Ix=Ixyz(1);Iy=Ixyz(2);Iz=Ixyz(3)
      Kxyz=(/1.12d0*Ix-0.12d0*Iz,Iy,Iz/)
      CALL xyz2lab(Kxyz,Kxyzwhite,Labs)
      Ls=Labs(1);as=Labs(2);bs=Labs(3)
!      L99d=325.22d0*log(1.d0+3.6d-3*Ls)
      L99d=scal_L99d*log(1.d0+3.6d-3*Ls)!exact normalisation to L=100
      e=as*c50+bs*s50
      f=1.14d0*(-as*s50+bs*c50)
!      WRITE(*,'(A,4(F16.8))') '**** a,b,e,f       =',as,bs,e,f
      G=sqrt(e*e+f*f)
      C99d=22.5d0*log(1.d0+0.06d0*G)
!---- Corrected arctan formula
!      IF (e>0.d0.AND.f>=0.d0) THEN
!         hef=atan(f/e)
!      ELSE IF (e==0.d0.AND.f>0.d0) THEN
!         hef=pi2
!      ELSE IF (e<0.d0) THEN
!         hef=pi+atan(f/e)
!      ELSE IF (e==0.d0.AND.f<0.d0) THEN
!         hef=1.5d0*pi
!      ELSE IF (e>0.d0.AND.f<=0.d0) THEN
!         hef=tau+atan(f/e)
!      ELSE!e==0 && f==0
!         hef=0.d0
!      ENDIF
      hef=atan2(f,e)!atan2 does quadrant checking
!      WRITE(*,'(A,4(F16.8))') '**** e,f,atan(f/e) =',e,f,hef
      h99d=hef+d50
      IF (h99d<0.d0) THEN
         h99d=h99d+tau
      ELSE IF (h99d>tau) THEN
         h99d=h99d-tau
      ENDIF
      a99d=C99d*cos(h99d)
      b99d=C99d*sin(h99d)
      Lab99d=(/L99d,a99d,b99d/)
      IF (degmode>=1) THEN
         LCh99d=(/L99d,C99d,h99d*r2d/)
      ELSE
         LCh99d=(/L99d,C99d,h99d/)
      ENDIF
      RETURN
    END SUBROUTINE XYZ2DIN99d

    SUBROUTINE LCh99d2XYZ(LCh99d,Ixyzwhite,Ixyz,degmode)
      IMPLICIT NONE
      INTEGER :: degmode
      REAL(KIND=CF_DOUBLE) :: scal_L99d=325.2194339489675351d0
      REAL(KIND=CF_DOUBLE) :: Ix,Iy,Iz,Kx,Ky,Kz,Ls,as,bs,L99d,C99d,h99d,a99d,b99d
      REAL(KIND=CF_DOUBLE) :: e,f,G,hef
      REAL(KIND=CF_DOUBLE) :: d50=0.87266462599716479d0
      REAL(KIND=CF_DOUBLE) :: c50=0.64278760968653933d0
      REAL(KIND=CF_DOUBLE) :: s50=0.76604444311897804d0
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Ixyz,Ixyzwhite,Kxyz,Kxyzwhite,Labs,LCh,Lab99d,LCh99d
      Ix=Ixyzwhite(1);Iy=Ixyzwhite(2);Iz=Ixyzwhite(3)
      Kxyzwhite=(/1.12d0*Ix-0.12d0*Iz,Iy,Iz/)
      L99d=LCh99d(1)
      C99d=LCh99d(2)
      h99d=LCh99d(3)
      IF (degmode>=1) THEN
         h99d=LCh99d(3)/r2d
      ELSE
         h99d=LCh99d(3)
      ENDIF
      G=(exp(C99d/22.5d0)-1.d0)/6.d-2
      e=G*cos(h99d-d50)
      f=G*sin(h99d-d50)
      as=e*c50-f/1.14d0*s50
      bs=e*s50+f/1.14d0*c50
!      Ls=(exp(L99d/325.22d0)-1.d0)/3.6d-3
      Ls=(exp(L99d/scal_L99d)-1.d0)/3.6d-3!exact normalisation to L=100
      Labs=(/Ls,as,bs/)
      CALL lab2xyz(Labs,Kxyz,Kxyzwhite)
      Kx=Kxyz(1);Ky=Kxyz(2);Kz=Kxyz(3)
      Iy=Ky;Iz=Kz
      Ix=(Kx+0.12d0*Kz)/1.12d0
      Ixyz=(/Ix,Iy,Iz/)
      RETURN
    END SUBROUTINE LCh99d2XYZ

    SUBROUTINE Lab99d2XYZ(Lab99d,Ixyzwhite,Ixyz)
      IMPLICIT NONE
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Ixyz,Ixyzwhite,Lab99d,LCh99d
      REAL(KIND=CF_DOUBLE) :: L99d,C99d,h99d,a99d,b99d,xdum
      L99d=Lab99d(1);a99d=Lab99d(2);b99d=Lab99d(3);
      C99d=sqrt(a99d**2+b99d**2)
      IF (C99d.gt.0.d0) THEN
         xdum=max(min(a99d/(C99d),1.d0),-1.d0)
      ELSE
         xdum=0.d0
      ENDIF
      h99d=acos(xdum)*sign(1.d0,b99d)
      LCh99d=(/L99d,C99d,h99d/)
      CALL LCh99d2XYZ(LCh99d,Ixyzwhite,Ixyz,0)
      RETURN
    END SUBROUTINE Lab99d2XYZ

    SUBROUTINE Lab2L99o(Lab,Lab99o,LCh99o,degmode)
      IMPLICIT NONE
      INTEGER :: degmode
      REAL(KIND=CF_DOUBLE) :: scal_L99o=303.6710054705098647d0
      REAL(KIND=CF_DOUBLE) :: d26=0.45378560551852569d0
      REAL(KIND=CF_DOUBLE) :: c26=0.89879404629916699d0
      REAL(KIND=CF_DOUBLE) :: s26=0.43837114678907742d0
      REAL(KIND=CF_DOUBLE) :: Ls,as,bs,L99o,a99o,b99o,C99o,h99o,eo,fo,G,hef
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Lab,Lab99o,LCh99o
      Ls=Lab(1); as=Lab(2); bs=Lab(3)
      L99o=scal_L99o/kE_DIN99*log(1.d0+3.9d-3*Ls)
      eo=as*c26+bs*s26
      fo=0.83d0*(-as*s26+bs*c26)
      G=sqrt(eo*eo+fo*fo)
      C99o=log(1.d0+0.075d0*G)/(0.0435d0*kE_DIN99*kCH_DIN99)
      hef=atan2(fo,eo)
      h99o=hef+d26
      IF (h99o<0.d0) THEN
         h99o=h99o+tau
      ELSE IF (h99o>tau) THEN
         h99o=h99o-tau
      ENDIF
      a99o=C99o*cos(h99o)
      b99o=C99o*sin(h99o)
      IF (degmode>=1) THEN
         LCh99o=(/L99o,C99o,h99o*r2d/)
      ELSE
         LCh99o=(/L99o,C99o,h99o/)
      ENDIF
      Lab99o=(/L99o,a99o,b99o/)
      RETURN
    END SUBROUTINE Lab2L99o

    SUBROUTINE LCh99o2Lab(LCh99o,Lab,degmode)
      IMPLICIT NONE
      INTEGER :: degmode
      REAL(KIND=CF_DOUBLE) :: scal_L99o=303.6710054705098647d0
      REAL(KIND=CF_DOUBLE) :: d26=0.45378560551852569d0
      REAL(KIND=CF_DOUBLE) :: c26=0.89879404629916699d0
      REAL(KIND=CF_DOUBLE) :: s26=0.43837114678907742d0
      REAL(KIND=CF_DOUBLE) :: Ls,as,bs,L99o,a99o,b99o,C99o,h99o,eo,fo,G,hef
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Lab,Lab99o,LCh99o
      L99o=LCh99o(1)
      C99o=LCh99o(2)
      IF (degmode>=1) THEN
         h99o=LCh99o(3)/r2d
      ELSE
         h99o=LCh99o(3)
      ENDIF
      Ls=(exp(kE_DIN99*L99o/scal_L99o)-1.d0)/3.9d-3
      hef=h99o-d26
      G=(exp(0.0435d0*kE_DIN99*kCH_DIN99*C99o)-1.d0)/0.075d0
      eo=G*cos(hef)
      fo=G*sin(hef)
      as=(eo*c26-(fo/0.83d0)*s26)
      bs=(eo*s26+(fo/0.83d0)*c26)
      Lab=(/Ls,as,bs/)
      RETURN
    END SUBROUTINE LCh99o2Lab

    SUBROUTINE Lab99o2Lab(Lab99o,Lab,degmode)
      IMPLICIT NONE
      INTEGER :: degmode
      REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Lab99o,Lab,LCh99o
      CALL lab2lch(Lab99o,LCh99o,1,degmode)
      CALL LCh99o2Lab(LCh99o,Lab,degmode)
    END SUBROUTINE Lab99o2Lab

END MODULE color_functions_extended
