!    Test RGB Table
!    Copyright (C) 2012 by Ingo Thies 
!
!    TestRGBtable is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 2 of the License, or
!    (at your option) any later version.
!
!    TestRGBtable is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with TestRGBtable. If not, see <http://www.gnu.org/licenses/>.
!
  PROGRAM testrgbtable
    USE color_functions_extended
    USE palette_modules
    IMPLICIT NONE
    INTEGER :: i,nt,degmode=1
    REAL(KIND=CF_DOUBLE) :: t,rgb_gamma
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: red_prim,grn_prim,blu_prim,wht_prim
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: urgb_ini,xyz_ini
    REAL(KIND=CF_DOUBLE),DIMENSION(0:3,0:nlevels_max) :: rgbtable
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: srgb,rgblin,Ixyz,xyz
    REAL(KIND=CF_DOUBLE),DIMENSION(3) :: Lab,Lab99o,Lab99d,LCh,LCh99o,LCh99d
    CHARACTER(LEN=30) :: infile,xyzfile,labfile,l99ofile,l99dfile,allfile,gpauxfile
    LOGICAL :: iscolorhelix=.FALSE.
    WRITE(*,*) 'Enter name of GP-RGB color table:'
    READ(*,*) infile
    iscolorhelix=trim(infile)=='colorhelix.rgb'
    IF (iscolorhelix) THEN
       WRITE(*,*) 'Reading auxiliary data from Colorhelix'
       OPEN(10,FILE='colorhelix-rgbdat.aux')
       READ(10,*) red_prim
       READ(10,*) grn_prim
       READ(10,*) blu_prim
       READ(10,*) wht_prim
       READ(10,*) XYZwhite
       READ(10,*) rgb_gamma
       READ(10,*) urgb_ini
       READ(10,*) xyz_ini
       CLOSE(10)
    ELSE
       WRITE(*,*) 'Using default (=sRGB) auxiliary data'
       red_prim=srgb_red
       grn_prim=srgb_grn
       blu_prim=srgb_blu
       wht_prim=srgb_wht
       XYZwhite=whitepoint_d65
       rgb_gamma=0.d0!force wrap_rgb_gamma to switch to sRGB
       xyz_ini=srgb_wht
       urgb_ini=(/1.d0,1.d0,1.d0/)
    ENDIF
    CALL getrgb(red_prim,grn_prim,blu_prim,XYZwhite,mat_xyz2rgb,mat_rgb2xyz)
    OPEN(11,FILE=trim(infile))
    CALL skipline(11)
    DO i=0,nlevels_max
       READ(11,*,END=11) rgbtable(0:3,i)
       nt=i
    ENDDO
11  CLOSE(11)
    gpauxfile='xrgbtable-gp.aux'
    xyzfile='xrgbtable.xyz'
    labfile='xrgbtable.lab'
    l99ofile='xrgbtable.l99o'
    l99dfile='xrgbtable.l99d'
    allfile='xrgbtable-all-cart.dat'
    OPEN(16,file=gpauxfile)
    OPEN(20,file=xyzfile)
    OPEN(21,file=labfile)
    OPEN(22,file=l99ofile)
    OPEN(23,file=l99dfile)
    OPEN(30,file=allfile)
    WRITE(16,'("XWP   =",F9.6,"; YWP   =",F9.6,"; ZWP   =",F9.6)') XYZwhite
    WRITE(16,'("xwhite=",F9.6,"; ywhite=",F9.6,"; zwhite=",F9.6)') wht_prim
    WRITE(16,'("xred  =",F9.6,"; yred  =",F9.6,"; zred  =",F9.6)') red_prim
    WRITE(16,'("xgreen=",F9.6,"; ygreen=",F9.6,"; zgreen=",F9.6)') grn_prim
    WRITE(16,'("xblue =",F9.6,"; yblue =",F9.6,"; zblue =",F9.6)') blu_prim
    IF (iscolorhelix) THEN
       WRITE(16,'(A)') '##########################'
       WRITE(16,'(A)') 'load ''colorhelix-gp.aux'''
       WRITE(16,'(A)') '##########################'
    ENDIF
    CLOSE(16)
    WRITE(20,1001) '#i','Level','R','G','B','X','Y','Z','x','y','z'           
    WRITE(21,1001) '#i','Level','R','G','B','L*','a*','b*','C*','h*'          
    WRITE(22,1001) '#i','Level','R','G','B','L99o','a99o','b99o','C99o','h99o'
    WRITE(23,1001) '#i','Level','R','G','B','L99d','a99d','b99d','C99d','h99d'
    WRITE(30,1001) '#1','2','3','4','5','6','7','8','9','10','11','12','13','14','15','16','17','18','19'
    WRITE(30,1001) '#i','Level','R','G','B','X','Y','Z', &
         'L*','a*','b*','L99o','a99o','b99o','L99d','a99d','b99d',&
         'x','y'
    DO i=0,nt
       t=rgbtable(0,i)
       srgb(1)=rgbtable(1,i)
       srgb(2)=rgbtable(2,i)
       srgb(3)=rgbtable(3,i)
!       write(*,*) 'rgb_gamma =',rgb_gamma
       CALL wrap_rgb_gamma(srgb,rgblin,rgb_gamma,-1)
       CALL operate_matrix3x3(mat_rgb2xyz,rgblin,Ixyz)
       IF (sum(Ixyz)==0.d0) THEN
          xyz=xyz_ini
       ELSE
          xyz=Ixyz/sum(Ixyz)
       ENDIF
       CALL xyz2lab(Ixyz,XYZwhite,Lab)
       CALL Lab2L99o(Lab,Lab99o,LCh99o,degmode)
       CALL lab2lch(Lab,LCh,1,degmode)
       CALL lab2lch(Lab99o,LCh99o,1,degmode)
       CALL XYZ2DIN99d(Ixyz,XYZwhite,Lab99d,LCh99d,degmode)
       WRITE(20,1000) i,t,srgb,Ixyz,xyz
       WRITE(21,1000) i,t,srgb,lab,lch(2:3)
       WRITE(22,1000) i,t,srgb,lab99o,lch99o(2:3)
       WRITE(23,1000) i,t,srgb,lab99d,lch99d(2:3)
       WRITE(30,1000) i,t,srgb,Ixyz,lab,lab99d,lab99o,xyz(1:2)
    ENDDO
    CLOSE(20)
    CLOSE(21)
    CLOSE(22)
    CLOSE(30)
1000 FORMAT(I4,1X,F14.8,2X,8(2X,3(2X,F16.8)))
1001 FORMAT(A4,1X,A14,  2X,8(2X,3(2X,A16)))
  END PROGRAM testrgbtable

  SUBROUTINE skipline(iufile)
    CHARACTER(LEN=100) :: line
    CHARACTER(LEN=1) :: first
    INTEGER :: iufile,ufile,nskip,i
    LOGICAL :: verbose=.FALSE.,goodline
!    verbose=.TRUE.
    ufile=abs(iufile)
    nskip=0
10  READ(ufile,'(a)',ERR=50,END=60) line
    DO i=1,100
       first=line(i:i)
       IF (first.ne. ' '.AND. first.ne.char(0)) EXIT
    ENDDO
    goodline=first.NE.'#' .AND. first.NE.'C' .AND. first.NE.'c' .AND.  &
             first.NE.';' .AND. first.NE.'%' .AND. first.NE.'!' .AND. line.NE.' '
    IF (goodline) THEN
       BACKSPACE ufile
       IF (verbose) THEN
          WRITE(*,*)
          WRITE(*,*) 'SKIPLINE: Skipped',nskip,' lines'
          WRITE(*,*)
       ENDIF
       RETURN
    ELSE
       nskip=nskip+1
       GOTO 10
    ENDIF

50  CONTINUE
    WRITE(*,*) ' nextline: error in reading line from unit ',ufile
    STOP
60  CONTINUE
    PRINT*,' nextline: end of file (may be regular)'
    RETURN
  END SUBROUTINE skipline
